
-- # import logic
-- # import id

infix 5 <
infix 8 |
infixl 12 +, -
infixl 14 *

-- Natural numbers

data Nat : Set where
  zero : Nat
  suc  : Nat -> Nat

sucInj (n,m:Nat) : suc n == suc m -> n == m
sucInj refl = refl

sucNotZero (n:Nat) : Not (suc n == zero)
sucNotZero ()

zeroNotSuc (n:Nat) : Not (zero == suc n)
zeroNotSuc ()

(+) : Nat -> Nat -> Nat
n + zero  = n
n + suc m = suc (n + m)

plusAssoc : (n,m,p : Nat) -> n + (m + p) == (n + m) + p
plusAssoc n m  zero   = refl
plusAssoc n m (suc p) = cong suc (plusAssoc n m p)

plusZero : (n:Nat) -> 0 + n == n
plusZero  zero	 = refl
plusZero (suc n) = cong suc (plusZero n)

plusSuc : (n,m:Nat) -> suc n + m == suc (n + m)
plusSuc n  zero	  = refl
plusSuc n (suc m) = cong suc (plusSuc n m)

plusCommute : (n,m : Nat) -> n + m == m + n
plusCommute  zero m   = plusZero m
plusCommute (suc n) m = fol-plugin(plusCommute n m, plusSuc n m)

(*) : Nat -> Nat -> Nat
zero  * m = zero
suc n * m = m + n * m

(|) : Nat -> Nat -> Prop
m | n = Exist $ \p -> m * p == n

(<) : Nat -> Nat -> Prop
n < zero  = False
n < suc m = n < m \/ n == m

(>) : Nat -> Nat -> Prop
n > m = m < n

(<=) : Nat -> Nat -> Prop
n <= m = n < m \/ n == m

zeroLessSuc : (n:Nat) -> zero < suc n
zeroLessSuc zero = orIR refl
zeroLessSuc (suc n) = orIL (zeroLessSuc n)

lessSucSuc : (n,m:Nat) -> suc n < suc m -> n < m
lessSucSuc n (suc m) (orIL lt) = orIL (lessSucSuc n m lt)
lessSucSuc n (suc m) (orIR eq) = fol-plugin(eq, Nat)

sucLessSuc : (n,m:Nat) -> n < m -> suc n < suc m
sucLessSuc n (suc m) (orIL lt) = orIL (sucLessSuc n m lt)
sucLessSuc n (suc m) (orIR eq) = orIR (cong suc eq)

lessSuc : (n,m:Nat) -> n < m -> suc n <= m
lessSuc n (suc m) (orIL lt) = orIL (sucLessSuc _ _ lt)
lessSuc n (suc m) (orIR eq) = orIR (cong suc eq)

(-) : Nat -> Nat -> Nat
n     - zero  = n
zero  - m     = zero
suc n - suc m = n - m

minusSame : (n:Nat) -> n - n == zero
minusSame zero = refl
minusSame (suc n) = minusSame n

minusSuc : (n,m:Nat) -> m <= n -> suc n - m == suc (n - m)
minusSuc  zero	  zero	 (orIR refl)	= refl
minusSuc (suc n) (suc n) (orIR refl)	= minusSuc n n (orIR refl)
minusSuc (suc n)  zero	 (orIL _)	= refl
minusSuc (suc n) (suc m) (orIL "sm<sn") = minusSuc n m (orIL $ lessSucSuc m n "sm<sn")

minusPlus : (n,m:Nat) -> m <= n -> n == n - m + m
minusPlus n n (orIR refl) = fol-plugin(minusSame, plusZero)
minusPlus (suc n) m (orIL "m<sn") =
  fol-plugin( minusPlus n m "m<sn"
	    , minusSuc n m "m<sn"
	    , minusSame
	    , plusZero, plusSuc
	    )

minusLess : (n,m:Nat) -> n > m -> 0 < m -> n - m < n
minusLess (suc n) (suc zero) "sn>1" _ = orIR refl
minusLess (suc n) (suc (suc m)) "sn>ssm" _ = orIL ih
  where
    ih : n - (suc m) < n
    ih = minusLess n (suc m) (lessSucSuc _ _ "sn>ssm") (zeroLessSuc _)

data Trich (n,m:Nat) : Prop where
  less	  : n <  m -> Trich n m
  equal	  : n == m -> Trich n m
  greater : n >  m -> Trich n m

trich : (n,m:Nat) -> Trich n m
trich zero     zero   = equal refl
trich zero    (suc m) = less (zeroLessSuc m)
trich (suc n)  zero   = greater (zeroLessSuc n)
trich (suc n) (suc m) with trich n m =>
      (suc n) (suc m) (less "n<m")    = less $ lessSuc n m "n<m"
      (suc n) (suc n) (equal refl)    = equal refl
      (suc n) (suc m) (greater "n>m") = greater $ lessSuc m n "n>m"

