-- Normalizing If-Expressions
-- Author:: Andreas Abel
-- Created :: 2005-07-06 
-- Source  :: IfExpressions.agda

-- Cut out data type NIfExpr

-- # import Property

data Bool : Set where
    true  : Bool
    false : Bool

if : (A : Set) -> Bool -> A -> A -> A
if A true  t e = t
if A false t e = e

boolCase : (P : Bool -> Prop) ->
	    P false -> P true -> (b:Bool) -> P b
boolCase p f t true = t
boolCase p f t false = f

-- Fix some set of variables and some valuation

A : Set
f : A -> Bool

-- If expressions over variables in some set A

data IfExp : Set where
    Bit : Bool -> IfExp
    Var : A -> IfExp
    If  : IfExp -> IfExp -> IfExp -> IfExp

-- Semantics

eval : IfExp -> Bool
eval (Bit b) = b
eval (Var a) = f a
eval (If c t e) = if Bool (eval c) (eval t) (eval e)

-- Normalization of if expressions

-- composition of normal expressions

nif : IfExp -> IfExp -> IfExp -> IfExp
nif (Bit true)  t e = t
nif (Bit false) t e = e
nif (Var x)     t e = If (Var x) t e
nif (If a b c)  t e = If a (nif b t e) (nif c t e)
  -- invariant:: a is Var x

-- normalization

norm : IfExp -> IfExp
norm (Bit b) = Bit b
norm (Var a) = Var a
norm (If c t e) = nif (norm c) (norm t) (norm e)

-- Soundness of normalization ----------------------------------------

-- Induction for IfExp

indIfExp : (P : IfExp -> Prop) ->
            ((b:Bool) -> P (Bit b)) ->
            ((a:A) -> P (Var a)) ->
            ((c,t,e : IfExp) -> P c -> P t -> P e -> P (If c t e)) ->
            (e : IfExp) -> P e
indIfExp P bit var if (Bit b) = bit b
indIfExp P bit var if (Var a) = var a
indIfExp P bit var if (If c t e)  = if c t e (indIfExp P bit var if c)
                                               (indIfExp P bit var if t)  
                                               (indIfExp P bit var if e)

-- Lemma:: rotation for if (imported from Bool)

Rot_if : (A:Set) -> (b:Bool) -> Prop
Rot_if A b = (ct,ce:Bool) -> (t,e:A) -> 
                (if A (if Bool b ct ce) t e) ==
                (if A b (if A ct t e) (if A ce t e))

rot_if : (A:Set) -> (b:Bool) -> Rot_if A b
rot_if A = boolCase (Rot_if A)
           (fol-plugin ()) 
           (fol-plugin ()) 

-- Lemma:: soundness of nif

{- Lemma:: eval (nif c t e) == eval (If c t e)
          (emb implicit)
          
Proof:: By induction on c.

Case NIf:: 
  LHS     eval (nif (If a ct ce) t e)
       == eval (If a (nif ct t e) (nif ce t e))
       == if (f a) (eval (nif ct t e)) (eval (nif ce t e))
  (IH) == if (f a) (eval (If ct t e)) (eval (If (ce t e))
       == if (f a) (if (eval ct) (eval t) (eval e))
                   (if (eval ce) (eval t) (eval e))
  RHS     eval (If (If a ct ce) t e)
       == if (eval (If a ct ce)) (eval t) (eval e)
       == if (if (f a) (eval ct) (eval ce)) (eval t) (eval e)
 rot_if== LHS
-}

Sound_nif : IfExp -> Prop
Sound_nif = \ c -> (t,e : IfExp) ->
                 (eval (If c t e))
              == (eval (nif c t e))

sound_nif  : (c : IfExp) -> Sound_nif c
sound_nif = indIfExp Sound_nif
                  (boolCase (\ b -> Sound_nif (Bit b))
                    fol-plugin()  -- just reflexivity
                    fol-plugin()) -- just reflexivity
                  fol-plugin()    -- just reflexivity
                  (\ c t e ihc iht ihe -> fol-plugin(ihc,iht,ihe,rot_if Bool))

--                  (\ a t e ih_t ih_e ->
--                     fol-plugin(ih_t,ih_e,rot_if Bool))

-- Theorem:: soundness of norm

Sound_norm : (e:IfExp) -> Prop
Sound_norm = \ e ->
                  (eval e) 
               == (eval (norm e))

sound_norm : (e:IfExp) -> Sound_norm e
sound_norm = indIfExp Sound_norm
               (boolCase (\ b -> Sound_norm (Bit b))
                    fol-plugin()  -- just reflexivity
                    fol-plugin()) -- just reflexivity
               (fol-plugin())     -- just reflexivity
               (\ c t e ihc iht ihe -> fol-plugin (ihc,iht,ihe,sound_nif))
