
-- # import Property
-- # import Code
-- # import Common

{------------------------------------------------------------------------
    Properties
 ------------------------------------------------------------------------}

all (a : Set) : (a -> Bool) -> List a -> Prop
all p []      = True
all p (x::xs)  = isTrue (p x) /\ all p xs

isOrdered (a : Set) : (a -> a -> Bool) -> List a -> Prop
isOrdered leq []      = True
isOrdered leq (x::xs)  = isOrdered leq xs /\ all (leq x) xs

infix 7 `elem`
elem (a : Set) : a -> List a -> Prop
elem x []     = False
elem x (y::ys) = x == y \/ elem x ys

infix 7 `subSet`
subSet (a : Set) : List a -> List a -> Prop
subSet [] ys = True
subSet (x::xs) ys = elem x ys /\ subSet xs ys

isPermutation (a : Set) : List a -> List a -> Prop
isPermutation xs ys =
  length xs == length ys  /\  xs `subSet` ys  /\  ys `subSet` xs

isTransitive (a : Set) : (a -> a -> Bool) -> Prop
isTransitive (#) = (x, y, z : a) ->
		   isTrue (x # y) ==> isTrue (y # z) ==> isTrue (x # z)

isTotal (a : Set) : (a -> a -> Bool) -> Prop
isTotal (#) = (x, y : a) -> isTrue (x # y) \/ isTrue (y # x)

{------------------------------------------------------------------------
    Helper proofs
 ------------------------------------------------------------------------}

-- Proofs about subSet

subSetCons (a : Set) : (z : a) -> (xs, ys : List a) ->
			 subSet xs ys ==> subSet xs (z::ys)
subSetCons z [] ys = \_ -> tt
subSetCons z (x::xs) ys = fol-plugin(subSetCons z xs ys)

subSetElem (a : Set) : (z : a) -> (xs, ys : List a) ->
			 elem z xs ==> subSet xs ys ==> elem z ys
subSetElem z [] ys	= fol-plugin()
subSetElem z (x::xs) ys	= fol-plugin(subSetElem z xs ys, el)
  where
    el : a -> List a -> Prop
    el = elem

subSetRefl (a : Set) : (xs : List a) -> xs `subSet` xs
subSetRefl []	  = tt
subSetRefl (x::xs) = fol-plugin(subSetCons x xs xs, subSetRefl xs)

subSetTrans (a : Set) : (xs, ys, zs : List a) ->
			  subSet xs ys ==> subSet ys zs ==> subSet xs zs
subSetTrans [] ys zs	  = fol-plugin()
subSetTrans (x::xs) ys zs  = fol-plugin(subSetTrans xs ys zs, subSetElem x)

-- Proofs about isPermutation

permCons (a : Set) : (z : a) -> (xs, ys : List a) ->
		       isPermutation xs ys ==> isPermutation (z::xs) (z::ys)
permCons z xs ys = fol-plugin(Nat, subSetCons z)

permTrans (a : Set) :
  (xs, ys, zs : List a) ->
  isPermutation xs ys ==> isPermutation ys zs ==> isPermutation xs zs
permTrans xs ys zs = fol-plugin(subSetTrans)

-- Proofs about all

-- allElem
allElem (a : Set) : (p : a -> Bool) -> (z : a) -> (xs : List a) ->
		      elem z xs ==> all p xs ==> isTrue (p z)
allElem p z []	    = fol-plugin()
allElem p z (x::xs)  = fol-plugin(allElem p z xs)

-- allEntails
allEntails (a : Set) : (p, q : a -> Bool) -> (xs : List a) ->
			 ((x : a) -> isTrue (p x) ==> isTrue (q x)) ->
			 all p xs ==> all q xs
allEntails p q [] pq	  = fol-plugin()
allEntails p q (x::xs) pq  = fol-plugin(pq, allEntails p q xs pq)

-- allSubSet
allSubSet (a : Set) :
    (p : a -> Bool) -> (xs, ys : List a) ->
    subSet ys xs ==> all p xs ==> all p ys
allSubSet p xs [] _ _		= tt
allSubSet p xs (y::ys) sub allxs = fol-plugin(py, allys)
  where

    py : isTrue (p y)
    py = allElem p y xs yInXs allxs
      where
	yInXs : elem y xs
	yInXs = fol-plugin(sub)

    allys : all p ys
    allys = fol-plugin(allSubSet p xs ys, sub, allxs, subSetCons)

-- allPermutation
allPermutation (a : Set) :
    (p : a -> Bool) -> (xs, ys : List a) ->
    isPermutation xs ys ==> all p xs ==> all p ys
allPermutation p xs ys = fol-plugin(allSubSet)

{------------------------------------------------------------------------
    Proofs
 ------------------------------------------------------------------------}

-- Permutations -----------------------------------------------------------

-- insertion permutes

insertPermutes (a : Set) :
    (leq : a -> a -> Bool) -> (x : a) -> (xs : List a) ->
    isPermutation (x::xs) (insert' leq x xs)
insertPermutes leq x []	  = fol-plugin()
insertPermutes leq x (y::ys) = fol-plugin(remLen, remSub1, remSub2)
  where
    ih = insertPermutes leq x ys

    lhs = x :: y :: ys
    rhs = insert' leq x (y::ys)

    elim = ifElim (leq x y) (x::y::ys) (y :: insert' leq x ys)

    remLen : length lhs == length rhs
    remLen = elim (\zs -> length lhs == length zs)
		  fol-plugin()
		  fol-plugin(ih)

    remSub1 : lhs `subSet` rhs
    remSub1 = fol-plugin(remX, remY, remYs)
      where
	remX : [x] `subSet` rhs
	remX = elim (subSet [x])
		    fol-plugin()
		    fol-plugin(ih)

	remY : [y] `subSet` rhs
	remY = elim (subSet [y]) fol-plugin() fol-plugin()

	remYs : ys `subSet` rhs
	remYs = elim (subSet ys)
		     fol-plugin(subSetRefl, subSetCons)
		     fol-plugin(subSetCons, ih)

    remSub2 : rhs `subSet` lhs
    remSub2 = elim (\zs -> subSet zs lhs)
		   fol-plugin(subSetRefl, subSetCons)
		   fol-plugin( subSetRefl, subSetCons, ih
			     , subSetTrans (insert' leq x ys) (x::ys) (x::y::ys)
			     )

-- sorting permutes

sortPermutes (a : Set) : (leq : a -> a -> Bool) -> (xs : List a) ->
			   isPermutation xs (sort' leq xs)
sortPermutes leq []	= fol-plugin()
sortPermutes leq (x::xs) = fol-plugin( freeze isPermutation
				    , insertPermutes leq
				    , sortPermutes leq xs
				    , permCons x
				    , permTrans (x::xs)
				    )

-- Ordering ---------------------------------------------------------------

-- insertion preserves ordering

insertOrdered (a : Set) :
    (leq : a -> a -> Bool) -> (z : a) -> (xs : List a) ->
    isTransitive leq ==> isTotal leq ==>
    isOrdered leq xs ==> isOrdered leq (insert' leq z xs)

insertOrdered leq z [] _ _ _ = fol-plugin()
insertOrdered leq z (x::xs) trans total ord_xxs =
    elim (\zs -> isOrdered leq zs)
	 rem1
	 rem2
  where
    ih : isOrdered leq (insert' leq z xs)
    ih = insertOrdered leq z xs trans total fol-plugin(ord_xxs)

    elim = ifElim (leq z x) (z::x::xs) (x :: insert' leq z xs)

    rem1 : isTrue (leq z x) -> isOrdered leq (z :: x :: xs)
    rem1 leqzx = fol-plugin(z_leq_xs, ord_xxs, leqzx)
      where
	z_leq_xs : all (leq z) xs
	z_leq_xs = allEntails (leq x) (leq z) xs i fol-plugin(ord_xxs)
	  where
	    i : (y : a) -> isTrue (leq x y) ==> isTrue (leq z y)
	    i y = fol-plugin(trans, leqzx)

    rem2 : isFalse (leq z x) -> isOrdered leq (x :: insert' leq z xs)
    rem2 nleqzx = fol-plugin(ih, rem)
      where
	rem : all (leq x) (insert' leq z xs)
	rem = fol-plugin(freeze isPermutation all, perm, rem2a, allPermutation)
	  where
	    perm : isPermutation (z::xs) (insert' leq z xs)
	    perm = fol-plugin(insertPermutes)

	    rem2a : all (leq x) (z::xs)
	    rem2a = fol-plugin(leqxz, ord_xxs)
	      where
		leqxz : isTrue (leq x z)
		leqxz = fol-plugin(total, nleqzx, Bool)

-- sort' orders lists
sortOrdered (a : Set) : (leq : a -> a -> Bool) -> (xs : List a) ->
			  isTransitive leq ==> isTotal leq ==>
			  isOrdered leq (sort' leq xs)
sortOrdered leq [] _ _		    = tt
sortOrdered leq (x::xs) trans total  =
  insertOrdered leq x (sort' leq xs) trans total (sortOrdered leq xs trans total)

-- Main Theorem
sortSorts (a : Set) : (leq : a -> a -> Bool) -> (xs : List a) ->
			isTransitive leq ==> isTotal leq ==>
			isPermutation xs (sort' leq xs) /\
			isOrdered leq (sort' leq xs)
sortSorts leq xs = fol-plugin(sortPermutes, sortOrdered)

