{-# OPTIONS -fglasgow-exts #-}

module TypeChecker.Monad.Errors where

import Control.Monad.Error
import Control.Monad.Reader
import Text.PrettyPrint

import Syntax.Position
import TypeChecker.Monad
import Utils.Error

-- Error functions --------------------------------------------------------

-- Note: don't export
getErrorMsg :: MonadReader Environment m => m ErrorMsg
getErrorMsg = liftM envErrorMsg ask

-- Note: don't export
updateErrorMsg f env = env { envErrorMsg = f $ envErrorMsg env }

sayWhen :: MonadReader Environment m => Doc -> m a -> m a
sayWhen w = local (updateErrorMsg $ \e -> e { errWhen = w })

sayWhere :: MonadReader Environment m => Range -> m a -> m a
sayWhere w = local (updateErrorMsg $ \e -> e { errWhere = w })

sayWhat :: MonadReader Environment m => Doc -> m a -> m a
sayWhat w = local (updateErrorMsg $ \e -> e { errWhat = w })

failure :: (MonadReader Environment m, MonadError ErrorMsg m) =>
	   (ErrorMsg -> ErrorMsg) -> m a
failure e = do	msg <- getErrorMsg
		throwError $ e msg

