
module Plugins.Equality where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Text.PrettyPrint
import Control.Monad       ( when )
import Control.Monad.Trans ( liftIO )

import Syntax.Internal
import Syntax.Values
import Syntax.Position
import Syntax.Pretty

import TypeChecker
import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Options
import TypeChecker.UnfoldConstants
--import TypeChecker.Abstraction
import TypeChecker.Values
import TypeChecker.ValueToTerm
import TypeChecker.Universes
import TypeChecker.Errors

import Utils.Error
import Utils.Monads
import Utils.Names

import Interface.Options

import Plugins

eq_Eq		= dummyName "Eq"
eq_congruence	= dummyName "congruence"
eq_symmetry	= dummyName "symmetry"

plugin :: SomePlugin
plugin = mkPlugin
    Plugin  { plName		= "eq"
            , plHelp            = show (fsep $ pwords "Does automatic symmetry and substitutivity in equality proofs.")
				    ++ "\n"
	    , plInit		= \_ -> return ()
	    , plFinal		= \_ _ -> return ()
	    , plCheck		= checkEquality
	    , plInfer		= \_ _ _ _ -> failure noInferenceError
	    , plOptions		= simpleOptions []
	    , plDefaultOptions	= ()
	    }

checkEquality :: () -> () -> Range -> [Term] -> Type -> TC (Value, ())
checkEquality _ _ _ _ _ = failure $ customError "eq-plugin() currently disabled"
{-
checkEquality _ _ r [proof] t =
    sayWhen (whenSearchingForProof proof t) $
    sayWhere r $
    do	(b,v,w)	    <- isEq t
	proof_t	    <- snd <$> inferType proof
	(a,v',w')   <- isEq proof_t
	let solutions =
		[ (False, f) | f <- abstract v' v, f w' === w ] ++
		[ (True,  g) | g <- abstract w' v, g v' === w ]
	case solutions of
	    []	    -> failure noSolutionFoundError
	    sol:_   ->
		-- If we find a solution we build the proof term
		-- and run it through the typechecker (just to be sure).
		do  e <- buildProof a b v' w' sol
		    v <- checkType e t

		    -- Debug printing
		    verbose <- optVerbose <$> commandLineOptions
		    when (verbose >= 2) $
			do  t' <- unfoldAllConstants $ unEl t
			    liftIO $ print $
				vcat[ fsep  [ text "+++ eq-plugin proved"
					    , nest 4 $ pretty t'
					    ]
				    , nest 4 $ text ":" <+> pretty e
				    ]

		    return (v, ())
    where
	buildProof a b x y (useSym, f) =
	    do	a'  <- valueToTerm (unEl a)
		b'  <- valueToTerm (unEl b)
		x'  <- valueToTerm x
		y'  <- valueToTerm y
		f'  <- mkLam a f
		let xy = proof
		    yx = sym `apps` [ a', x', y', xy ]
		return $
		    if useSym
			then mkProof a' b' f' y' x' yx
			else mkProof a' b' f' x' y' xy
	    where
		cong	= Def noRange eq_congruence
		sym	= Def noRange eq_symmetry
		apps	= foldl App
		mkProof t u e x y xy
		    | isId a f	= xy
		    | otherwise	= cong `apps` [ t, u, e, x, y, xy ]

	isId a f = x === f x
	    where x = vId $ dummyName "<any>"

	mkLam v f =
	    do	x <- fresh "x"
		b <- valueToTerm (f $ vId x)
		return $ Lam noRange (VarP x) b

	isEq t =
	    do	v' <- unfoldConstants (unEl t)
		case vAppView v' of
		    VAVDef c _ [a,v,w]
			| c == eq_Eq    -> return (el vsSet a,v,w)
		    _			-> failure $ onlyEqualityProofsError (unEl t)
checkEquality _ _ _ _ _ =
    failure wrongNumberOfArgumentsError
-}

{-
testAbstraction [e1, e2] =
    do	inferType e1
	inferType e2
	w <- evaluateI e1
	v <- evaluateI e2
	liftIO $ print $
		    vcat[ fsep  [ text "Abstracting"
				, nest 2 $ pretty w
				, text "from"
				, nest 2 $ pretty v
				]
			, nest 2    $ vcat
				    $ map (pretty . mkLam)
				    $ abstract w v
			]
    where
	mkLam :: (Value -> Value) -> Value
	mkLam f = VLam (Cl "x" f) vSet
testAbstraction _ = return ()
-}

whenSearchingForProof p v =
    fsep $ pwords "searching for proof of" ++ [pretty v] ++
	   pwords "using" ++ [pretty p]

noInferenceError :: ErrorFun
noInferenceError e =
    e	{ errWhat   = fsep $ pwords "type inference failed"
	, errWhy    = fsep $ pwords "the type of an automatic equality proof cannot be inferred"
	}

onlyEqualityProofsError :: Value -> ErrorFun
onlyEqualityProofsError v e =
    e	{ errWhat   = fsep $ pwords "don't know what to do with" ++ [pretty v]
	, errWhy    = fsep $ pwords "it is not an equality"
	}

noSolutionFoundError :: ErrorFun
noSolutionFoundError e =
    e	{ errWhat   = fsep $ pwords "proof failed"
	, errWhy    = fsep $ pwords $
			"it wasn't a simple application of symmetry " ++
			"and/or congruence (or there is a bug in " ++
			"the proof search)"
	}

wrongNumberOfArgumentsError :: ErrorFun
wrongNumberOfArgumentsError e =
    e	{ errWhat   = fsep $ pwords "expected a single argument to eq-plugin"
	}

