{-# OPTIONS -fglasgow-exts #-}

module Plugins.FOL.Arguments
    ( parseArguments
    , Arguments(..)
    , AllowIncludes(..)
    ) where

import Data.Monoid
import Syntax.Internal

data Arguments = Args { premisses   :: [Term]
		      , freeze	    :: [Name]
		      , includes    :: [String]
		      }

data AllowIncludes = NoIncludes | WithIncludes
    deriving Eq

instance Monoid Arguments where
    mempty = Args [] [] []
    mappend (Args xs ys zs) (Args xs' ys' zs') =
	Args (mappend xs xs')
	     (mappend ys ys')
	     (mappend zs zs')

parseArguments :: AllowIncludes -> [Term] -> Arguments
parseArguments allowIncludes = mconcat . map arg
    where
	arg t = case nameAppView t of
		    NameArg (Name _ "freeze" _) ts
			| all isName ts	-> mempty { freeze = map getName ts }
		    NameArg (Name _ "include" _) [t]
			| allowIncludes == WithIncludes
			  , isName t
			  , '\"':x <- nameString $ getName t
			    -> mempty { includes = [init x] }
		    _	    -> mempty { premisses = [t] }

data NameAppView = NameArg Name [Term]
		 | Other Term

nameAppView (Var _ x)	= NameArg x []
nameAppView (Def _ x)	= NameArg x []
nameAppView (Con _ x)	= NameArg x []
nameAppView (Data _ x)	= NameArg x []
nameAppView (App s t)	= nameAppView s `app` t
    where
	app (NameArg x ts) t = NameArg x (ts ++ [t])
	app (Other s) t	     = Other (App s t)
nameAppView t		= Other t

getName (Var _ x)   = x
getName (Con _ x)   = x
getName (Data _ x)  = x
getName (Def _ x)   = x
getName (Sig _ x)   = x
getName t	    = error $ "impossible: Plugins.FOL.Arguments.getName"

isName (Var _ _)    = True
isName (Con _ _)    = True
isName (Data _ _)   = True
isName (Def _ _)    = True
isName (Sig _ _)    = True
isName _	    = False

