{-| Structural beta-eta-equality and universe-subtyping for values.
-}

module TypeChecker.Equality where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Control.Monad
import Control.Monad.Trans
import Data.List

import Syntax.Values
import Syntax.Internal

import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Options
import TypeChecker.Monad.MetaVars
import TypeChecker.Monad.Context
import TypeChecker.Monad.Signatures
import TypeChecker.Monad.Statistics
import TypeChecker.UnfoldConstants  ( unfoldConstants )
import TypeChecker.Values
import TypeChecker.Values.Refresh
import TypeChecker.Universes
--import TypeChecker.Sorts	    (unifySort, lessEqSort)
import TypeChecker.Errors
import TypeChecker.Abstractions
import TypeChecker.MetaVars
import TypeChecker.Context

import Utils.Names		    ( fresh )
import Utils.Monads

-- Pattern unification ----------------------------------------------------

patternUnification :: MetaRef -> [Value] -> ValueAppView -> TC ()
patternUnification m ps u
    | not (validParameters ps) = postponeConstraint
    | otherwise = flip catchPatternViolation postponeConstraint $ do
    incStat "patternUnification"
    case u of

	VAVMeta m' qs
	    | m == m'   ->
		do  -- The meta should take the same number of arguments
		    unless (length ps == length qs) $ failure $ customError "bad number of args"
		    -- Keep only the parameters that are the same
		    let vs = [ vBId n | (n, (p, q)) <- zip [0..] $ reverse
								 $ zip ps qs
				      , sameBId p q
			     ]
		    -- Instantiate
		    w <- createClosedMeta
		    m =: lamAbstraction (length ps) (vApps w vs)

	_	->
	    do  let v = unAppView u
-- 		    liftIO $ putStrLn $ "occursCheck " ++ show (unAppView $ VAVMeta m t ps) ++ " in " ++ show v
		occursCheck m ps v
-- 		    liftIO $ putStrLn $ "result: " ++ show v
		let lookupPar n ps = lup 0 n $ reverse ps
			where
			    lup _ _ [] = Nothing
			    lup i n (p : ps) =
				case vView p of
				    VBId m
					| n == m	-> Just $ vBId i
					| otherwise	-> lup (i + 1) n ps
				    _	-> error "impossible: TypeChecker.Equality.lookupPar"

		len <- nofBoundVars <$> getContext
		let inst =  [ maybe (vBId (-1)) id (lookupPar n ps)
			    | n <- [0..len]
			    ]
-- 		    liftIO $ putStrLn $ "instantiating: " ++
-- 					    show (buildMetaLam (instantiate v inst) ps)
		m =: lamAbstraction (length ps) (instantiate v inst)
-- 		    liftIO $ putStrLn "Done"
    where
	postponeConstraint =
	    postpone m (vApps (vMeta m) ps === unAppView u)

-- First order unficiation ------------------------------------------------

-- | The first value is on the form @vApps (vMeta m) vs@.
firstOrderUnification :: Value -> Value -> TC ()
firstOrderUnification v w =
    case (vView v, vView w) of
	(VApp v1 v2, VApp w1 w2) -> do	v1 === w1
					v2 === w2
	(VApp _ _, _)	-> failure $ notAtAllEqualError v w
	(VMeta m, VMeta m')
	    | m == m'	->  return ()
	(VMeta m, _)	->  do	occursCheck m w
				verbose 12 $ liftIO $ putStrLn $ "first order instantiation to " ++ show w
				m =: w
	_		-> error "impossible: TypeChecker.Equality.firstOrderUnification"
    where
	occursCheck m w = applyEverywhere occurs w
	    where
		occurs v =
		    case vView v of
			VMeta m' | m == m'  -> failure $ customError "occurs check"
			_		    -> return v

-- Equality ---------------------------------------------------------------

(===) :: Value -> Value -> TC ()
v1 === v2
    | v1 == v2	= return () --incStat "eq match"
    | otherwise	=
    do	--incStat "eq mismatch"
	[v1',v2'] <- mapM (unfoldConstants <.> refreshSpine) [v1,v2]
	verbose 12 $ liftIO $ putStrLn $ "conversion: " ++ show v1' ++ " == " ++ show v2'
	vAppView v1' ==* vAppView v2'


VAVLam f ==* VAVLam g = f ==@ g

VAVLam f ==* u =
    freshParameter oops $ \v ->
	absApp f v === vApply (unAppView u) v
    where
	oops = error "conversion checking doesn't know what to add to the context"

u ==* VAVLam f =
    freshParameter oops $ \v ->
	vApply (unAppView u) v === absApp f v
    where
	oops = error "conversion checking doesn't know what to add to the context"

VAVPair v w ==* VAVPair v' w' =
    do	v === v'
	w === w'

VAVPair v w ==* u =
    do	v === vProject (unAppView u) VFst
	w === vProject (unAppView u) VSnd

u ==* VAVPair v w =
    do	vProject (unAppView u) VFst === v
	vProject (unAppView u) VSnd === w

VAVPi v f ==* VAVPi w g =
    do	unEl v === unEl w
	fmap unEl f ==@ fmap unEl g

VAVSigma v f ==* VAVSigma w g =
	do  unEl v === unEl w
	    fmap unEl f ==@ fmap unEl g

VAVSort s1 ==* VAVSort s2 =
--     ifM withUniversePolymorphism
-- 	(unifySort s1 s2)
	(unless (s1 == s2) $ failure $ sortsNotEqualError s1 s2)

VAVTop ==* VAVTop = return ()

VAVId DontCare _ ==* _		      = failure dontCareNotEqualError
_		 ==* VAVId DontCare _ = failure dontCareNotEqualError

VAVId x vs ==* VAVId y ws
    | x == y    = zipWithM_ (===) vs ws
    | otherwise = failure $ variablesNotEqualError x y

VAVBId n vs ==* VAVBId m ws
    | n == m    = zipWithM_ (===) vs ws
    | otherwise = failure $ variablesNotEqualError (dummyName $ "@" ++ show n) (dummyName $ "@" ++ show m)
    -- TODO: error msg

VAVDef x _ vs ==* VAVDef y _ ws
    | x == y    =
	do  verbose 12 $ liftIO $ putStrLn $ "Same def. Checking spines: " ++ show vs ++ " == " ++ show ws
	    zipWithM_ (===) vs ws
    | otherwise = failure $ variablesNotEqualError x y

VAVCon x _ _ _ vs ==* VAVCon y _ _ _ ws
    | x == y    = zipWithM_ (===) vs ws
    | otherwise = failure $ variablesNotEqualError x y

VAVProj v p vs ==* VAVProj w q ws
    | p == q	= zipWithM_ (===) (v:vs) (w:ws)

v@(VAVMeta m ps) ==* u =
    ifM isFirstOrderUnification
	(firstOrderUnification (unAppView v) (unAppView u))
	(patternUnification m ps u)

u ==* v@(VAVMeta _ _) = v ==* u

VAVStruct xs ==* VAVStruct ys =
    zipWithM_ (===) (map (snd . snd) xs) (map (snd . snd) ys)

u@(VAVStruct xs) ==* v = u ==* VAVStruct ys
    where
	ys = [ (x,(h,unAppView v `vProj` VField h x)) | (x,(h,_)) <- xs ]

u ==* v@(VAVStruct _) = v ==* u

u ==* v = failure $ notAtAllEqualError (unAppView u) (unAppView v)

(==@) :: Abs Value -> Abs Value -> TC ()
cl ==@ cl'  =
    do	x <- fresh $ absName cl
	extendContext x (Abstract oops) $
	    absBody cl === absBody cl'
    where
	oops = error $ "conversion checking doesn't know what to add to the context"

data Tree = Leaf Value | Node Tree Tree

-- | Non-empty list. This is used when inferring the type
--   of a multiple clause definition or a case. Only works
--   for a limited class of definitions, of course.
--   (Namely the empty class.)
allEqual :: [Value] -> TC Value
allEqual = treeEqual . mkTree . map Leaf
    where
	mkTree [t]	= t
	mkTree (l:r:ts) = mkTree (Node l r : ts)
	mkTree _ = error "impossible: TypeChecker.Equality.mkTree"

	treeEqual (Leaf v) = return v
	treeEqual (Node t1 t2) =
	    do	v1 <- treeEqual t1
		v2 <- treeEqual t2
		v1 === v2
		return v1

-- Subtyping --------------------------------------------------------------

(=<=) :: Type -> Type -> TC ()
v =<= w
    | v == w	= return () --incStat "leq match"
    | otherwise	=
    do	--incStat "leq mismatch"
	v' <- unfoldConstants (unEl v)
	w' <- unfoldConstants (unEl w)
	vAppView v' =<=* vAppView w'

VAVSort s1 =<=* VAVSort s2 =
--     ifM withUniversePolymorphism
-- 	(lessEqSort s1 s2)
	(unless (s1 <= s2) $ failure $ sortsNotLeqError s1 s2)

v@(VAVDef x _ vs) =<=* w@(VAVDef y _ ws) =
    do	zipWithM_ (===) vs ws
	sub <- isSubsignature x y
	unless (x == y || sub) $ failure $ variablesNotSubtypeError x y

-- ? Alternative: v === w, theoretical implications ?
VAVPi v f =<=* VAVPi w g =
    do	w =<= v
	f =<=@ g

VAVSigma v f =<=* VAVSigma w g =
	do  v =<= w
	    f =<=@ g

v =<=* w = unAppView v === unAppView w

(=<=@) :: Abs Type -> Abs Type -> TC ()
cl =<=@ cl' =
    do	x <- fresh $ absName cl
	extendContext x (Abstract oops) $
	    absBody cl =<= absBody cl'
    where
	oops = error $ "subtype check doesn't know what type to add to the context"

