{-# OPTIONS -fglasgow-exts #-}

module TypeChecker.Patterns where

import Syntax.Position
import Syntax.Internal
import Syntax.Values
import TypeChecker.Values
import TypeChecker.Monad
import TypeChecker.Abstractions

import Utils.Monads

patternVars :: Pattern -> [Name]
patternVars (WildP _)		= []
patternVars (AbsurdP _)		= []
patternVars (VarP x)		= [x]
patternVars (PairP _ p q)	= patternVars p ++ patternVars q
patternVars (ConP _ _ hs ps)	= concatMap patternVars (hs ++ ps)

nPatternVars :: PatternValue -> Int
nPatternVars p = case p of
    VWildP     -> 0
    VAbsurdP   -> 0
    VIdP       -> 1
    VPairP p q -> nPatternVars p + nPatternVars q
    VConP _ ps -> sum $ map nPatternVars ps

patternInstantiation :: Pattern -> [Value]
patternInstantiation = inst []
    where
	inst r (AbsurdP _)	= []
	inst r (WildP _)	= [vProjs (vBId 0) r]
	inst r (VarP x)		= [vProjs (vBId 0) r]
	inst r (PairP _ p q)	= inst (VSnd:r) q ++ inst (VFst:r) p
	inst r (ConP _ _ _ _)	= error "impossible: TypeChecker.Patterns.patternInstantiation (ConP _ _ _ _)"

patName (WildP _)	= "_"
patName (AbsurdP _)	= "()"
patName (VarP x)	= nameString x
patName (PairP _ p q)	= patName p ++ patName q
patName (ConP _ _ _ _)	= error "impossible: TypeChecker.Patterns.patName (ConP _ _ _ _)"

-- | Assumed to be linear.
patternValue :: Pattern -> PatternValue
patternValue (VarP _)	      = VIdP
patternValue (WildP _)	      = VWildP
patternValue (AbsurdP _)      = VAbsurdP
patternValue (PairP _ p q)    = VPairP (patternValue p) (patternValue q)
patternValue (ConP _ x ps qs) = VConP x $ map patternValue $ ps ++ qs

