
-- # import nat

-- Fin
data Fin : Nat -> Set where
  fzero (n : Nat) : Fin (suc n)
  fsuc  (n : Nat) : Fin n -> Fin (suc n)

-- Types
data Typ : Set where
  nat   : Typ
  (==>) : Typ -> Typ -> Typ

infixr 5 ==>

-- Contexts
data Ctx : Nat -> Set where
  empty		: Ctx 0
  (&) (n : Nat) : Ctx n -> Typ -> Ctx (suc n)

infixl 10 &

-- Variables
data Var' : (n : Nat) -> Ctx n -> Typ -> Set where
  vzero	(n : Nat)(ctx : Ctx n)(t : Typ) : Var' (suc n) (ctx & t) t
  vsuc  (n : Nat)(ctx : Ctx n)(s, t : Typ) :
       Var' n ctx t -> Var' (suc n) (ctx & s) t

Var (n : Nat) : Ctx n -> Typ -> Set
Var ctx t = Var' _ ctx t

-- Terms
data Term' : (n : Nat) -> Ctx n -> Typ -> Set where
  var (n : Nat)(ctx : Ctx n)(t : Typ) : Var ctx t -> Term' n ctx t

  (@) (n : Nat)(ctx : Ctx n)(s, t : Typ) :
    Term' n ctx (s ==> t) -> Term' n ctx s -> Term' n ctx t

  lam (n : Nat)(ctx : Ctx n)(s, t : Typ) :
    Term' (suc n) (ctx & s) t -> Term' n ctx (s ==> t)

infixl 15 @

Term (n : Nat) : Ctx n -> Typ -> Set
Term ctx t = Term' _ ctx t

-- Thinning
insert (n : Nat) : Fin (suc n) -> Typ -> Ctx n -> Ctx (suc n)
insert fzero	t ctx	    = ctx & t
insert (fsuc i) t (ctx & s) = insert i t ctx & s

thinVar (n : Nat)(s,t : Typ)(ctx : Ctx n) :
	(i : Fin (suc n)) -> Var ctx t -> Var (insert i s ctx) t
thinVar fzero v		  = vsuc v
thinVar (fsuc i) vzero	  = vzero
thinVar (fsuc i) (vsuc v) = vsuc (thinVar i v)

data ThinVarView (n : Nat)(i : Fin (suc n))(ctx : Ctx n) :
      (s, t : Typ) -> Var (insert i s ctx) t -> Set where

  hit  (s : Typ)(v : Var (insert i s ctx) s) : ThinVarView n i ctx s s v
  miss (s, t : Typ) : (v : Var ctx t) -> ThinVarView n i ctx s t (thinVar i v)

thinVarView (n : Nat)(s, t : Typ) :
	    (i : Fin (suc n)) -> (ctx : Ctx n) -> (v : Var (insert i s ctx) t) ->
	    ThinVarView n i ctx s t v
thinVarView fzero ctx vzero		  = hit
thinVarView fzero ctx (vsuc v)		  = miss v
thinVarView (fsuc i) (ctx & t') vzero	  = miss vzero
thinVarView (fsuc i) (ctx & t') (vsuc v) with thinVarView i ctx v =>
  _ _ _ (vsuc v)	      hit     = hit
  _ _ _ (vsuc (thinVar _ v)) (miss v) = miss (vsuc v)

-- = thinVarViewSuc v (thinVarView i ctx v)

thinTerm (n : Nat)(s, t : Typ)(ctx : Ctx n) :
	 (i : Fin (suc n)) -> Term ctx t -> Term (insert i s ctx) t
thinTerm i (var v) = var (thinVar i v)
thinTerm i (f @ e) = thinTerm i f @ thinTerm i e
thinTerm i (lam e) = lam (thinTerm (fsuc i) e)

-- Substitution
substitute (n : Nat)(s, t : Typ)(ctx : Ctx n) :
      (i : Fin (suc n)) ->
      Term ctx s -> Term (insert i s ctx) t -> Term ctx t
substitute i e (var v) with thinVarView _ _ v =>
  _ i e (var  v)	       hit      = e
  _ i e (var (thinVar i v))   (miss v)  = var v
substitute i e (e1 @ e2) = substitute i e e1 @ substitute i e e2
substitute i e (lam e')  = lam (substitute (fsuc i) (thinTerm fzero e) e')

-- Evaluation
reduce (n : Nat)(t : Typ)(ctx : Ctx n) : Term ctx t -> Term ctx t
reduce (lam e1 @ e2)  = substitute fzero e2 e1
reduce (lam e)	      = lam (reduce e)
reduce (var v @ e2)   = var v @ reduce e2
reduce (var v)	      = var v
reduce (e1 @ e2 @ e3) = reduce (e1 @ e2) @ e3

reduceN (n : Nat)(t : Typ)(ctx : Ctx n) : Nat -> Term ctx t -> Term ctx t
reduceN 0 e	  = e
reduceN (suc m) e = reduce (reduceN m e)

idNat (n : Nat)(ctx : Ctx n) : Term ctx (nat ==> nat)
idNat = lam (var vzero)

tmK (n : Nat)(ctx : Ctx n)(s,t : Typ) : Term ctx (s ==> t ==> s)
tmK = lam (lam (var (vsuc (vzero))))

tmS (n : Nat)(ctx : Ctx n)(s,t,u : Typ) : Term ctx ((s ==> t ==> u) ==> (s ==> t) ==> s ==> u)
tmS = lam (lam (lam (var (vsuc (vsuc vzero)) @ var vzero @ (var (vsuc vzero) @ var vzero))))

test : Nat -> Term empty (nat ==> nat)
test n = reduceN n (tmS @ tmK @ tmK {_} {_} {_} {nat})

foo : Nat -> Nat
foo n = n

