-- an example from Michael Beeson

-- # import Property

-- some logical axioms

infix 14 *
infix 10 +

Ex : (A : Set) -> (A -> Prop) -> Prop
witness : (A : Set) -> (P : A -> Prop) -> (x : A) -> P x -> Ex A P
elimEx : (A : Set) -> (P : A -> Prop) -> Ex A P -> (C : Prop) -> ((x : A)-> P x ==> C) -> C

-- natural numbers

data Nat : Set where
 zero : Nat
 succ : Nat -> Nat

indNat : (P : Nat -> Prop) -> P zero -> ((n : Nat) -> P n -> P (succ n)) -> (x : Nat) -> P x
indNat P a g zero = a
indNat P a g (succ n) = g n (indNat P a g n)

-- general definitions

isRing : (X : Set) -> (X -> X -> X) -> (X -> X -> X) -> (X -> X) -> X -> X -> Prop

isRing R (+) (*) minus Zero One =
 (x : R) -> (y : R) -> (z : R) ->
  (   (x + y) == (y + x)
   /\ (x + Zero) == x
   /\ (x + (minus x)) == Zero
   /\ (x + (y + z)) == ((x + y) + z)
   /\ (x * (y + z)) == ((x * y) + (x * z))
   /\ ((y + z)* x) == ((y * x) + (z * x))
   /\ (x * One) == x
   /\ (One * x) == x
   /\ (x * (y * z)) == ((x * y) * z)
  )

isIntegral : (X : Set) -> (X -> X -> X) -> (X -> X -> X) -> (X -> X) -> X -> X -> Prop

isIntegral R (+) (*) minus Zero One =
     isRing R (+) (*) minus Zero One 
 /\ ((x : R) -> (y : R) -> (x * y) == Zero ==> (x == Zero \/ y == Zero))

-- the example

R : Set
(+) : R -> R -> R
(*) : R -> R -> R
minus : R -> R 
Zero : R
One : R

axR : isIntegral  R (+) (*) minus Zero One 

power : Nat -> R -> R
power zero x = One
power (succ n) x = (power n x) * x

isZero : R -> Prop
isZero x = x == Zero


lemZero : (x : R) -> isZero (Zero * x) 
lemZero x =  fol-plugin(axR)

lemOneZero : (x : R) -> One == Zero ==> isZero x
lemOneZero x = fol-plugin(axR,lemZero)

prop : R -> Nat -> Prop
prop x n = isZero (power n x) ==> isZero x

lem1 : (x : R) -> (n : Nat) -> prop x n
lem1 x =
 let
  base : prop x zero
  base = fol-plugin(lemOneZero)
  step : (n : Nat) -> prop x n -> prop x (succ n)
  step = \ n -> \ h1 -> fol-plugin(axR,h1)
 in
  indNat (prop x) base step

isNilpotent : R -> Prop
isNilpotent x = Ex Nat (\ n -> isZero (power n x))

thm : (x : R) -> isNilpotent x -> isZero x
thm x h = elimEx Nat (\ n -> isZero (power n x)) h (isZero x) (lem1 x)

     
